@@ -87,7 +87,9 @@ class HyperparameterTuning:
         self.target_variables = target_variables
         self.device_type = device_type
         if self.device_type is None:
-            self.device_type = "gpu" if torch.cuda.is_available() else "cpu"
+            # Use new device detection with MPS support
+            from .utils import get_optimal_device
+            _, self.device_type = get_optimal_device()
         self.surv_event_var = surv_event_var
         self.surv_time_var = surv_time_var
         self.batch_variables = batch_variables
@@ -140,9 +142,11 @@ class HyperparameterTuning:
     
     def setup_trainer(self, params, current_step, total_steps, full_train = False):
         # Configure callbacks and trainer for the current fold
-        mycallbacks = [self.progress_bar]
+        mycallbacks = []
         if self.plot_losses:
             mycallbacks.append(LiveLossPlot(hyperparams=params, current_step=current_step, total_steps=total_steps))
+        else:
+            mycallbacks.append(self.progress_bar)
         # when training on a full dataset; no cross-validation or no validation splits; 
         # we don't do early stopping
         early_stop_callback = None
@@ -152,7 +156,7 @@ class HyperparameterTuning:
     
         trainer = pl.Trainer(
             #deterministic = True, 
-            precision = '16-mixed', # mixed precision training 
+            #precision = '16-mixed', # mixed precision training 
             max_epochs=int(params['epochs']),
             gradient_clip_val=1.0,  
             gradient_clip_algorithm='norm',
