@@ -10,9 +10,10 @@ import os, argparse
 from scipy import stats
 from functools import reduce
 
-from captum.attr import IntegratedGradients
+from captum.attr import IntegratedGradients, GradientShap
 
 from ..modules import *
+from ..utils import to_device_safe
 
 class DirectPred(pl.LightningModule):
     """
@@ -64,13 +65,21 @@ class DirectPred(pl.LightningModule):
                 hidden_dim=int(self.input_dims[i] * self.config['hidden_dim_factor']),
                 output_dim=self.config['latent_dim']) for i in range(len(self.layers))])
 
+        if len(self.input_dims) > 1:
+            self.fusion_block = nn.Linear(
+                in_features=self.config['latent_dim'] * len(self.layers),
+                out_features=self.config['latent_dim']
+            )
+        else:
+            self.fusion_block = None
+        
         self.MLPs = nn.ModuleDict()  # using ModuleDict to store multiple MLPs
         for var in self.variables:
             if self.variable_types[var] == 'numerical':
                 num_class = 1
             else:
                 num_class = len(np.unique(self.ann[var]))
-            self.MLPs[var] = MLP(input_dim=self.config['latent_dim'] * len(self.layers),
+            self.MLPs[var] = MLP(input_dim=self.config['latent_dim'],
                                  hidden_dim=self.config['supervisor_hidden_dim'],
                                  output_dim=num_class)
 
@@ -89,10 +98,12 @@ class DirectPred(pl.LightningModule):
         for i, x in enumerate(x_list):
             embeddings_list.append(self.encoders[i](x))
         embeddings_concat = torch.cat(embeddings_list, dim=1)
+        # if multiple embeddings, fuse them 
+        embeddings = self.fusion_block(embeddings_concat) if self.fusion_block else embeddings_concat
 
         outputs = {}
         for var, mlp in self.MLPs.items():
-            outputs[var] = mlp(embeddings_concat)
+            outputs[var] = mlp(embeddings)
         return outputs  
     
     
@@ -259,7 +270,8 @@ class DirectPred(pl.LightningModule):
             dict: Predicted values mapped by target variable names.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
 
         # Create a DataLoader with a practical batch size
@@ -270,20 +282,21 @@ class DirectPred(pl.LightningModule):
         # Process each batch
         for batch in dataloader:
             dat, y_dict, samples = batch
-            x_list = [dat[x].to(device) for x in dat.keys()]  # Prepare the data batch for processing
+            x_list = [to_device_safe(dat[x], device) for x in dat.keys()]  # Prepare the data batch for processing
 
             # Perform the forward pass
             outputs = self.forward(x_list)
 
             # Collect predictions for each variable
             for var in self.variables:
-                y_pred = outputs[var].detach().cpu().numpy()  # Move outputs back to CPU and convert to numpy
+                logits = outputs[var].detach().cpu()  # Raw model outputs (logits)
+                
                 if dataset.variable_types[var] == 'categorical':
-                    predictions[var].extend(np.argmax(y_pred, axis=1))
+                    probs = torch.softmax(logits, dim=1).numpy() # class probabilities between 0 and 1
+                    predictions[var].extend(probs)
                 else:
-                    predictions[var].extend(y_pred)
-
-        # Convert lists to arrays if necessary, depending on the downstream use-case
+                    predictions[var].extend(logits.numpy()) # return raw output for regression problems
+        # Convert lists to arrays 
         predictions = {var: np.array(predictions[var]) for var in predictions}
 
         return predictions
@@ -299,7 +312,8 @@ class DirectPred(pl.LightningModule):
             pd.DataFrame: DataFrame containing the transformed data.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
 
         dataloader = DataLoader(dataset, batch_size=64, shuffle=False)  # Adjust the batch size as needed
@@ -313,13 +327,16 @@ class DirectPred(pl.LightningModule):
             batch_embeddings = []
             # Process each input matrix with its corresponding Encoder
             for i, x in enumerate(dat.values()):
-                x = x.to(device)  # Move data to GPU
+                x = to_device_safe(x, device)  # Move data to GPU
                 encoded_x = self.encoders[i](x)  # Transform data using the corresponding encoder
                 batch_embeddings.append(encoded_x)
             
             # Concatenate all embeddings from the current batch
             embeddings_batch_concat = torch.cat(batch_embeddings, dim=1)
-            embeddings_list.append(embeddings_batch_concat.detach().cpu())  # Move tensor back to CPU and detach
+            # if multiple embeddings, fuse them 
+            embeddings_batch = self.fusion_block(embeddings_batch_concat) if self.fusion_block else embeddings_batch_concat
+
+            embeddings_list.append(embeddings_batch.detach().cpu())  # Move tensor back to CPU and detach
             sample_names.extend(samples)  # Collect sample names
 
         # Concatenate all batch embeddings into one tensor
@@ -331,11 +348,11 @@ class DirectPred(pl.LightningModule):
                                      columns=[f"E{dim}" for dim in range(embeddings_concat.shape[1])])
         return embeddings_df
         
-    # Adaptor forward function for captum integrated gradients. 
+    # Adaptor forward function for captum integrated gradients or gradient shap 
     def forward_target(self, *args):
         input_data = list(args[:-2])  # one or more tensors (one per omics layer)
         target_var = args[-2]  # target variable of interest
-        steps = args[-1]  # number of steps for IntegratedGradients().attribute 
+        steps = args[-1]  # number of steps/samples for IntegratedGradients().attribute or GradientShap.attribute 
         outputs_list = []
         for i in range(steps):
             # get list of tensors for each step into a list of tensors
@@ -344,32 +361,37 @@ class DirectPred(pl.LightningModule):
             outputs_list.append(out[target_var])
         return torch.cat(outputs_list, dim = 0)
 
-    def compute_feature_importance(self, dataset, target_var, steps=5, batch_size = 64):
+    def compute_feature_importance(self, dataset, target_var, method="IntegratedGradients", steps_or_samples=5, batch_size=64):
         """
-        Computes the feature importance for each variable in the dataset using the Integrated Gradients method.
-        This method measures the importance of each feature by attributing the prediction output to each input feature.
-    
+        Computes the feature importance for each variable in the dataset using either Integrated Gradients or Gradient SHAP.
+
         Args:
             dataset: The dataset object containing the features and data.
             target_var (str): The target variable for which feature importance is calculated.
-            steps (int, optional): The number of steps to use for integrated gradients approximation. Defaults to 5.
+            method (str, optional): The attribution method to use ("IntegratedGradients" or "GradientShap").
+                                    Defaults to "IntegratedGradients".
+            steps_or_samples (int, optional): Number of steps for Integrated Gradients or samples for Gradient SHAP.
+                                              Defaults to 5.
             batch_size (int, optional): The size of the batch to process the dataset. Defaults to 64.
-    
+
         Returns:
             pd.DataFrame: A DataFrame containing feature importances across different variables and data modalities.
-                          Columns include 'target_variable', 'target_class', 'target_class_label', 'layer', 'name',
-                          and 'importance'.
-    
-        This function adjusts the device setting based on the availability of GPUs and performs the computation using
-        Integrated Gradients. It processes batches of data, aggregates results across batches, and formats the output
-        into a readable DataFrame which is then stored in the model's attribute for later use or analysis.
         """
-        device = torch.device("cuda" if self.device_type == 'gpu' and torch.cuda.is_available() else 'cpu')
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)
 
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
-        ig = IntegratedGradients(self.forward_target)
 
+        # Choose the attribution method dynamically
+        if method == "IntegratedGradients":
+            explainer = IntegratedGradients(self.forward_target)
+        elif method == "GradientShap":
+            explainer = GradientShap(self.forward_target)
+        else:
+            raise ValueError(f"Unsupported method '{method}'. Choose 'IntegratedGradients' or 'GradientShap'.")
+
+        # Handle target class (numerical vs categorical)
         if dataset.variable_types[target_var] == 'numerical':
             num_class = 1
         else:
@@ -378,53 +400,63 @@ class DirectPred(pl.LightningModule):
         aggregated_attributions = [[] for _ in range(num_class)]
         for batch in dataloader:
             dat, _, _ = batch
-            x_list = [dat[x].to(device) for x in dat.keys()]
+            x_list = [to_device_safe(dat[x], device) for x in dat.keys()]
             input_data = tuple([data.unsqueeze(0).requires_grad_() for data in x_list])
-            baseline = tuple(torch.zeros_like(x) for x in input_data)
+            
+            if method == 'IntegratedGradients':
+                baseline = tuple(torch.zeros_like(x) for x in input_data)
+            elif method == 'GradientShap': # provide multiple baselines for Gr.Shap
+                baseline = tuple(
+                    torch.cat([torch.zeros_like(x) for _ in range(steps_or_samples)], dim=0)
+                    for x in input_data
+                )
             if num_class == 1:
                 # returns a tuple of tensors (one per data modality)
-                attributions = ig.attribute(input_data, baseline, 
-                                             additional_forward_args=(target_var, steps), 
-                                             n_steps=steps)
+                if method == 'IntegratedGradients':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_steps=steps_or_samples)
+                elif method == 'GradientShap':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_samples=steps_or_samples)
                 aggregated_attributions[0].append(attributions)
             else:
                 for target_class in range(num_class):
                     # returns a tuple of tensors (one per data modality)
-                    attributions = ig.attribute(input_data, baseline, 
-                                                 additional_forward_args=(target_var, steps), 
-                                                 target=target_class, n_steps=steps)
+                    if method == 'IntegratedGradients':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_steps=steps_or_samples)
+                    elif method == 'GradientShap':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_samples=steps_or_samples)
                     aggregated_attributions[target_class].append(attributions)
-
-        # For each target class and for each data modality/layer, concatenate attributions accross batches 
+        # Post-process attributions
         layers = list(dataset.dat.keys())
         num_layers = len(layers)
-        processed_attributions = [] 
-        # Process each class
+        processed_attributions = []
         for class_idx in range(len(aggregated_attributions)):
             class_attr = aggregated_attributions[class_idx]
             layer_attributions = []
-            # Process each layer within the class
             for layer_idx in range(num_layers):
-                # Extract all batch tensors for this layer across all batches for the current class
                 layer_tensors = [batch_attr[layer_idx] for batch_attr in class_attr]
-                # Concatenate tensors along the batch dimension
                 attr_concat = torch.cat(layer_tensors, dim=1)
                 layer_attributions.append(attr_concat)
             processed_attributions.append(layer_attributions)
-        
-        # compute absolute importance and move to cpu 
+
         abs_attr = [[torch.abs(a).cpu() for a in attr_class] for attr_class in processed_attributions]
-        # average over samples 
         imp = [[a.mean(dim=1) for a in attr_class] for attr_class in abs_attr]
-        # move the model also back to cpu (if not already on cpu)
         self.to('cpu')
 
-        # combine into a single data frame
+        # Combine results into a DataFrame
         df_list = []
         for i in range(num_class):
             for j in range(len(layers)):
                 features = dataset.features[layers[j]]
-                # Ensure tensors are already on CPU before converting to numpy
                 importances = imp[i][j][0].detach().numpy()
                 target_class_label = dataset.label_mappings[target_var].get(i) if target_var in dataset.label_mappings else ''
                 df_list.append(pd.DataFrame({'target_variable': target_var, 
@@ -434,6 +466,5 @@ class DirectPred(pl.LightningModule):
                                              'name': features, 
                                              'importance': importances}))    
         df_imp = pd.concat(df_list, ignore_index=True)
-        # save the computed scores in the model
         self.feature_importances[target_var] = df_imp
-    
+
