@@ -10,7 +10,7 @@ import lightning as pl
 
 from torch.utils.data import DataLoader
 
-from captum.attr import IntegratedGradients
+from captum.attr import IntegratedGradients, GradientShap
 
 from ..modules import MLP, cox_ph_loss, flexGCN
 
@@ -76,8 +76,8 @@ class GNN(pl.LightningModule):
             self.target_variables = self.target_variables + [self.surv_event_var]
         self.batch_variables = batch_variables
         self.variables = self.target_variables + self.batch_variables if self.batch_variables else self.target_variables
-        self.variable_types = dataset.multiomic_dataset.variable_types 
-        self.ann = dataset.multiomic_dataset.ann 
+        self.variable_types = getattr(dataset, 'multiomic_dataset', dataset).variable_types 
+        self.ann = getattr(dataset, 'multiomic_dataset', dataset).ann 
         self.edge_index = dataset.edge_index
         
         self.feature_importances = {}
@@ -86,7 +86,8 @@ class GNN(pl.LightningModule):
         self.device_type = device_type 
         self.gnn_conv_type = gnn_conv_type
         
-        device = torch.device("cuda" if self.device_type == 'gpu' and torch.cuda.is_available() else 'cpu')
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.edge_index = self.edge_index.to(device) # edge index is re-used across samples, so we keep it in device
                 
         if self.use_loss_weighting:
@@ -294,7 +295,8 @@ class GNN(pl.LightningModule):
             dict: Predictions mapped by target variable names.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
 
         # Create a DataLoader with a practical batch size
@@ -308,16 +310,15 @@ class GNN(pl.LightningModule):
             x = x.to(device)  # Move data to GPU
 
             outputs = self.forward(x, edge_index)
-
-            # Collect predictions for each variable
             for var in self.variables:
-                y_pred = outputs[var].detach().cpu().numpy()  # Move outputs back to CPU and convert to numpy
-                if self.variable_types[var] == "categorical":
-                    predictions[var].extend(np.argmax(y_pred, axis=1))
-                else:
-                    predictions[var].extend(y_pred)
+                logits = outputs[var].detach().cpu()  # Raw model outputs (logits)
 
-        # Convert lists to arrays if necessary, depending on the downstream use-case
+                if dataset.variable_types[var] == 'categorical':
+                    probs = torch.softmax(logits, dim=1).numpy() # class probabilities between 0 and 1
+                    predictions[var].extend(probs)
+                else:
+                    predictions[var].extend(logits.numpy()) # return raw output for regression problems
+        # Convert lists to arrays 
         predictions = {var: np.array(predictions[var]) for var in predictions}
 
         return predictions
@@ -333,7 +334,8 @@ class GNN(pl.LightningModule):
             pd.DataFrame: DataFrame containing the transformed data.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
         edge_index = dataset.edge_index.to(device)  # Move edge_index to GPU
 
@@ -375,34 +377,38 @@ class GNN(pl.LightningModule):
         return torch.cat(outputs_list, dim = 0)
 
         
-    def compute_feature_importance(self, dataset, target_var, steps=5, batch_size = 64):
+    def compute_feature_importance(self, dataset, target_var, method="IntegratedGradients", steps_or_samples=5, batch_size=64):
         """
-        Computes the feature importance for each variable in the dataset using the Integrated Gradients method.
-        This method measures the importance of each feature by attributing the prediction output to each input feature.
-    
+        Computes the feature importance for each variable in the dataset using either Integrated Gradients or Gradient SHAP.
+
         Args:
-            dataset: The dataset object containing the features and data (MultiOmicDatasetNW object).
+            dataset: The dataset object containing the features and data.
             target_var (str): The target variable for which feature importance is calculated.
-            steps (int, optional): The number of steps to use for integrated gradients approximation. Defaults to 5.
+            method (str, optional): The attribution method to use ("IntegratedGradients" or "GradientShap").
+                                    Defaults to "IntegratedGradients".
+            steps_or_samples (int, optional): Number of steps for Integrated Gradients or samples for Gradient SHAP.
+                                              Defaults to 5.
             batch_size (int, optional): The size of the batch to process the dataset. Defaults to 64.
-    
+
         Returns:
             pd.DataFrame: A DataFrame containing feature importances across different variables and data modalities.
-                          Columns include 'target_variable', 'target_class', 'target_class_label', 'layer', 'name',
-                          and 'importance'.
-    
-        This function adjusts the device setting based on the availability of GPUs and performs the computation using
-        Integrated Gradients. It processes batches of data, aggregates results across batches, and formats the output
-        into a readable DataFrame which is then stored in the model's attribute for later use or analysis.
         """
         def bytes_to_gb(bytes):
             return bytes / 1024 ** 2
-        device = torch.device("cuda" if self.device_type == 'gpu' and torch.cuda.is_available() else 'cpu')
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)
         self.dataset_edge_index = dataset.edge_index.to(device)
 
         dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
-        ig = IntegratedGradients(self.forward_target)
+        
+        # Choose the attribution method dynamically
+        if method == "IntegratedGradients":
+            explainer = IntegratedGradients(self.forward_target)
+        elif method == "GradientShap":
+            explainer = GradientShap(self.forward_target)
+        else:
+            raise ValueError(f"Unsupported method '{method}'. Choose 'IntegratedGradients' or 'GradientShap'.")
 
         if dataset.variable_types[target_var] == 'numerical':
             num_class = 1
@@ -418,18 +424,34 @@ class GNN(pl.LightningModule):
             input_data = x.unsqueeze(0).requires_grad_().to(device)
             baseline = torch.zeros_like(input_data)
             
+            if method == 'IntegratedGradients':
+                baseline = torch.zeros_like(input_data)
+            elif method == 'GradientShap': # provide multiple baselines for Gr.Shap
+                baseline = torch.cat([torch.zeros_like(input_data) for _ in range(steps_or_samples)], dim=0)
+                    
             if num_class == 1:
                 # returns a tuple of tensors (one per data modality)
-                attributions = ig.attribute( input_data, baseline, 
-                                             additional_forward_args=(target_var, steps), 
-                                             n_steps=steps)
+                if method == 'IntegratedGradients':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_steps=steps_or_samples)
+                elif method == 'GradientShap':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_samples=steps_or_samples)
                 aggregated_attributions[0].append(attributions)
             else:
                 for target_class in range(num_class):
-                    # returns a tuple of tensors (one per data modality)
-                    attributions = ig.attribute( input_data, baseline, 
-                                                 additional_forward_args=(target_var, steps), 
-                                                 target=target_class, n_steps=steps)
+                    if method == 'IntegratedGradients':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_steps=steps_or_samples)
+                    elif method == 'GradientShap':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_samples=steps_or_samples)
                     aggregated_attributions[target_class].append(attributions)
         # For each target class concatenate node attributions accross batches 
         processed_attributions = [] 
@@ -451,7 +473,7 @@ class GNN(pl.LightningModule):
 
 
         df_list = []
-        layers = list(dataset.multiomic_dataset.dat.keys())
+        layers = list(getattr(dataset, 'multiomic_dataset', dataset).dat.keys())
         for i in range(num_class):
             features = dataset.common_features
             target_class_label = dataset.label_mappings[target_var].get(i) if target_var in dataset.label_mappings else ''
