@@ -11,7 +11,7 @@ import numpy as np
 import lightning as pl
 from scipy import stats
 
-from captum.attr import IntegratedGradients
+from captum.attr import IntegratedGradients, GradientShap
 
 from ..modules import *
 
@@ -73,15 +73,16 @@ class supervised_vae(pl.LightningModule):
         input_dims = [len(dataset.features[layers[i]]) for i in range(len(layers))]
         # create a list of Encoder instances for separately encoding each omics layer
         self.encoders = nn.ModuleList([Encoder(input_dims[i], 
-                                               # define hidden_dim size as a factor of input_dim  
-                                               [int(input_dims[i] * config['hidden_dim_factor'])], 
+                                               # define hidden_dim size as a factor of input_dim; keep at least 2 units
+                                               [max(int(input_dims[i] * config['hidden_dim_factor']), 2)], 
                                                config['latent_dim']) for i in range(len(layers))])
         # Fully connected layers for concatenated means and log_vars
         self.FC_mean = nn.Linear(len(layers) * config['latent_dim'], config['latent_dim'])
         self.FC_log_var = nn.Linear(len(layers) * config['latent_dim'], config['latent_dim'])
         # list of decoders to decode each omics layer separately 
         self.decoders = nn.ModuleList([Decoder(config['latent_dim'], 
-                                               [int(input_dims[i] * config['hidden_dim_factor'])], 
+                                               # define hidden_dim size as a factor of input_dim; keep at least 2 units
+                                               [max(int(input_dims[i] * config['hidden_dim_factor']),2)], 
                                                input_dims[i]) for i in range(len(layers))])
 
         # define supervisor heads
@@ -344,7 +345,8 @@ class supervised_vae(pl.LightningModule):
             pd.DataFrame: Transformed dataset as a pandas DataFrame.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
 
         dataloader = DataLoader(dataset, batch_size=64, shuffle=False)  # Adjust the batch size as needed
@@ -383,7 +385,8 @@ class supervised_vae(pl.LightningModule):
             dict: Predicted values mapped by target variable names.
         """
         self.eval()  # Set the model to evaluation mode
-        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)  # Move the model to the appropriate device
 
         dataloader = DataLoader(dataset, batch_size=64, shuffle=False)  # Adjust the batch size as needed
@@ -400,13 +403,14 @@ class supervised_vae(pl.LightningModule):
 
             # Collect predictions for each variable
             for var in self.variables:
-                y_pred = outputs[var].detach().cpu().numpy()  # Move outputs back to CPU and convert to numpy
+                logits = outputs[var].detach().cpu()  # Raw model outputs (logits)
+                
                 if dataset.variable_types[var] == 'categorical':
-                    predictions[var].extend(np.argmax(y_pred, axis=1))
+                    probs = torch.softmax(logits, dim=1).numpy() # class probabilities between 0 and 1
+                    predictions[var].extend(probs)
                 else:
-                    predictions[var].extend(y_pred)
-
-        # Convert lists to arrays if necessary, depending on the downstream use-case
+                    predictions[var].extend(logits.numpy()) # return raw output for regression problems
+        # Convert lists to arrays 
         predictions = {var: np.array(predictions[var]) for var in predictions}
 
         return predictions
@@ -480,32 +484,34 @@ class supervised_vae(pl.LightningModule):
             outputs_list.append(outputs[target_var])
         return torch.cat(outputs_list, dim = 0)
         
-    def compute_feature_importance(self, dataset, target_var, steps = 5, batch_size = 64):
+    def compute_feature_importance(self, dataset, target_var, method="IntegratedGradients", steps_or_samples=5, batch_size=64):
         """
-        Computes the feature importance for each variable in the dataset using the Integrated Gradients method.
-        This method measures the importance of each feature by attributing the prediction output to each input feature.
-    
+        Computes the feature importance for each variable in the dataset using either Integrated Gradients or Gradient SHAP.
+
         Args:
             dataset: The dataset object containing the features and data.
             target_var (str): The target variable for which feature importance is calculated.
-            steps (int, optional): The number of steps to use for integrated gradients approximation. Defaults to 5.
+            method (str, optional): The attribution method to use ("IntegratedGradients" or "GradientShap").
+                                    Defaults to "IntegratedGradients".
+            steps_or_samples (int, optional): Number of steps for Integrated Gradients or samples for Gradient SHAP.
+                                              Defaults to 5.
             batch_size (int, optional): The size of the batch to process the dataset. Defaults to 64.
-    
+
         Returns:
             pd.DataFrame: A DataFrame containing feature importances across different variables and data modalities.
-                          Columns include 'target_variable', 'target_class', 'target_class_label', 'layer', 'name',
-                          and 'importance'.
-    
-        This function adjusts the device setting based on the availability of GPUs and performs the computation using
-        Integrated Gradients. It processes batches of data, aggregates results across batches, and formats the output
-        into a readable DataFrame which is then stored in the model's attribute for later use or analysis.
         """
-        device = torch.device("cuda" if self.device_type == 'gpu' and torch.cuda.is_available() else 'cpu')
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)
         
         print("[INFO] Computing feature importance for variable:",target_var,"on device:",device)
-        # Initialize the Integrated Gradients method
-        ig = IntegratedGradients(self.forward_target)
+        # Choose the attribution method dynamically
+        if method == "IntegratedGradients":
+            explainer = IntegratedGradients(self.forward_target)
+        elif method == "GradientShap":
+            explainer = GradientShap(self.forward_target)
+        else:
+            raise ValueError(f"Unsupported method '{method}'. Choose 'IntegratedGradients' or 'GradientShap'.")
 
         # Get the number of classes for the target variable
         if self.dataset.variable_types[target_var] == 'numerical':
@@ -521,19 +527,37 @@ class supervised_vae(pl.LightningModule):
             dat, _, _ = batch
             x_list = [dat[x].to(device) for x in dat.keys()]
             input_data = tuple([data.unsqueeze(0).requires_grad_() for data in x_list])
-            baseline = tuple(torch.zeros_like(x) for x in input_data)
+            
+            if method == 'IntegratedGradients':
+                baseline = tuple(torch.zeros_like(x) for x in input_data)
+            elif method == 'GradientShap': # provide multiple baselines for Gr.Shap
+                baseline = tuple(
+                    torch.cat([torch.zeros_like(x) for _ in range(steps_or_samples)], dim=0)
+                    for x in input_data
+                )
+            
             if num_class == 1:
-                # returns a tuple of tensors (one per data modality)
-                attributions = ig.attribute(input_data, baseline, 
-                                             additional_forward_args=(target_var, steps), 
-                                             n_steps=steps)
+                if method == 'IntegratedGradients':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_steps=steps_or_samples)
+                elif method == 'GradientShap':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(target_var, steps_or_samples), 
+                                                 n_samples=steps_or_samples)
                 aggregated_attributions[0].append(attributions)
             else:
                 for target_class in range(num_class):
-                    # returns a tuple of tensors (one per data modality)
-                    attributions = ig.attribute(input_data, baseline, 
-                                                 additional_forward_args=(target_var, steps), 
-                                                 target=target_class, n_steps=steps)
+                    if method == 'IntegratedGradients':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_steps=steps_or_samples)
+                    elif method == 'GradientShap':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                           additional_forward_args=(target_var, steps_or_samples), 
+                                                           target=target_class,
+                                                           n_samples=steps_or_samples)
                     aggregated_attributions[target_class].append(attributions)
 
         # For each target class and for each data modality/layer, concatenate attributions accross batches 
