@@ -14,7 +14,7 @@ import lightning as pl
 from ..modules import *
 from ..data import TripletMultiOmicDataset
 
-from captum.attr import IntegratedGradients
+from captum.attr import IntegratedGradients, GradientShap
 
 
 
@@ -78,6 +78,15 @@ class MultiTripletNetwork(pl.LightningModule):
                 hidden_dim=int(self.input_dims[i] * self.config['hidden_dim_factor']),
                 output_dim=self.config['latent_dim']) for i in range(len(self.layers))])
         
+        if len(self.input_dims) > 1:
+            self.fusion_block = nn.Linear(
+                in_features=self.config['latent_dim'] * len(self.layers),
+                out_features=self.config['latent_dim']
+            )
+        else:
+            self.fusion_block = None
+
+        
         # define supervisor heads for both target and batch variables 
         self.MLPs = nn.ModuleDict() # using ModuleDict to store multiple MLPs
         for var in self.variables:
@@ -85,7 +94,7 @@ class MultiTripletNetwork(pl.LightningModule):
                 num_class = 1
             else:
                 num_class = len(np.unique(self.ann[var]))
-            self.MLPs[var] = MLP(input_dim=self.config['latent_dim'] * len(self.layers),
+            self.MLPs[var] = MLP(input_dim=self.config['latent_dim'],
                                  hidden_dim=self.config['supervisor_hidden_dim'],
                                  output_dim=num_class)
     
@@ -96,7 +105,9 @@ class MultiTripletNetwork(pl.LightningModule):
         for i, x in enumerate(x_list):
             embeddings_list.append(self.encoders[i](x))
         embeddings_concat = torch.cat(embeddings_list, dim=1)
-        return embeddings_concat
+        # if multiple embeddings, fuse them 
+        embeddings = self.fusion_block(embeddings_concat) if self.fusion_block else embeddings_concat
+        return embeddings
         
     def forward(self, anchor, positive, negative):
         """
@@ -341,13 +352,19 @@ class MultiTripletNetwork(pl.LightningModule):
             outputs[var] = mlp(anchor_embedding)
         
         # get predictions from the mlp outputs for each var
-        predictions = {}
+        predictions = {var: [] for var in self.variables}  # Initialize prediction storage
+        # Collect predictions for each variable
         for var in self.variables:
-            y_pred = outputs[var].detach().numpy()
-            if self.variable_types[var] == 'categorical':
-                predictions[var] = np.argmax(y_pred, axis=1)
+            logits = outputs[var].detach().cpu()  # Raw model outputs (logits)
+
+            if dataset.variable_types[var] == 'categorical':
+                probs = torch.softmax(logits, dim=1).numpy() # class probabilities between 0 and 1
+                predictions[var].extend(probs)
             else:
-                predictions[var] = y_pred
+                predictions[var].extend(logits.numpy()) # return raw output for regression problems
+        # Convert lists to arrays 
+        predictions = {var: np.array(predictions[var]) for var in predictions}
+        
         return predictions
 
     
@@ -370,28 +387,25 @@ class MultiTripletNetwork(pl.LightningModule):
             outputs_list.append(outputs[target_var])
         return torch.cat(outputs_list, dim = 0)
         
-    def compute_feature_importance(self, dataset, target_var, steps = 5, batch_size = 64):
+    def compute_feature_importance(self, dataset, target_var, method="IntegratedGradients", steps_or_samples=5, batch_size=64):
         """
-        Computes the feature importance for each variable in the dataset using the Integrated Gradients method.
-        This method measures the importance of each feature by attributing the prediction output to each input feature.
-    
+        Computes the feature importance for each variable in the dataset using either Integrated Gradients or Gradient SHAP.
+
         Args:
             dataset: The dataset object containing the features and data.
             target_var (str): The target variable for which feature importance is calculated.
-            steps (int, optional): The number of steps to use for integrated gradients approximation. Defaults to 5.
+            method (str, optional): The attribution method to use ("IntegratedGradients" or "GradientShap").
+                                    Defaults to "IntegratedGradients".
+            steps_or_samples (int, optional): Number of steps for Integrated Gradients or samples for Gradient SHAP.
+                                              Defaults to 5.
             batch_size (int, optional): The size of the batch to process the dataset. Defaults to 64.
-    
+
         Returns:
             pd.DataFrame: A DataFrame containing feature importances across different variables and data modalities.
-                          Columns include 'target_variable', 'target_class', 'target_class_label', 'layer', 'name',
-                          and 'importance'.
-    
-        This function adjusts the device setting based on the availability of GPUs and performs the computation using
-        Integrated Gradients. It processes batches of data, aggregates results across batches, and formats the output
-        into a readable DataFrame which is then stored in the model's attribute for later use or analysis.
         """
 
-        device = torch.device("cuda" if self.device_type == 'gpu' and torch.cuda.is_available() else 'cpu')
+        from ..utils import create_device_from_string
+        device = create_device_from_string(self.device_type if hasattr(self, 'device_type') and self.device_type else 'auto')
         self.to(device)
 
         print("[INFO] Computing feature importance for variable:",target_var,"on device:",device)
@@ -399,8 +413,14 @@ class MultiTripletNetwork(pl.LightningModule):
         # define data loader 
         triplet_dataset = TripletMultiOmicDataset(dataset, self.main_var)
         dataloader = DataLoader(triplet_dataset, batch_size=batch_size, shuffle=False)
-        # Initialize the Integrated Gradients method
-        ig = IntegratedGradients(self.forward_target)
+        
+        # Choose the attribution method dynamically
+        if method == "IntegratedGradients":
+            explainer = IntegratedGradients(self.forward_target)
+        elif method == "GradientShap":
+            explainer = GradientShap(self.forward_target)
+        else:
+            raise ValueError(f"Unsupported method '{method}'. Choose 'IntegratedGradients' or 'GradientShap'.")
 
         if self.variable_types[target_var] == 'numerical':
             num_class = 1
@@ -431,22 +451,36 @@ class MultiTripletNetwork(pl.LightningModule):
             layer_sizes = [anchor[i].shape[1] for i in range(len(anchor))] 
 
             # Define a baseline 
-            baseline = torch.zeros_like(input_data)       
-
+            if method == 'IntegratedGradients':
+                baseline = torch.zeros_like(input_data) 
+            elif method == 'GradientShap': # provide multiple baselines for Gr.Shap
+                baseline =  torch.cat([torch.zeros_like(input_data) for _ in range(steps_or_samples)], dim=0)
+                    
             if num_class == 1:
-                # returns a tuple of tensors (one per data modality)
-                attributions = ig.attribute(input_data, baseline, 
-                                             additional_forward_args=(layer_sizes, target_var, steps), 
-                                             n_steps=steps)
-                attributions = attributions.split(layer_sizes, dim = 3)
+                
+                if method == 'IntegratedGradients':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(layer_sizes, target_var, steps_or_samples), 
+                                                 n_steps=steps_or_samples)
+                    attributions = attributions.split(layer_sizes, dim = 3)
+                elif method == 'GradientShape':
+                    attributions = explainer.attribute(input_data, baseline, 
+                                                 additional_forward_args=(layer_sizes, target_var, steps_or_samples), 
+                                                 n_samples=steps_or_samples)
+                    attributions = attributions.split(layer_sizes, dim = 3)
                 aggregated_attributions[0].append(attributions)
             else:
                 for target_class in range(num_class):
-                    # returns a tuple of tensors (one per data modality)
-                    attributions = ig.attribute(input_data, baseline, 
-                                                 additional_forward_args=(layer_sizes, target_var, steps), 
-                                                 target=target_class, n_steps=steps)
-                    attributions = attributions.split(layer_sizes, dim = 3)
+                    if method == 'IntegratedGradients':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                     additional_forward_args=(layer_sizes, target_var, steps_or_samples), 
+                                                     target=target_class, n_steps=steps_or_samples)
+                        attributions = attributions.split(layer_sizes, dim = 3)
+                    elif method == 'GradientShap':
+                        attributions = explainer.attribute(input_data, baseline, 
+                                                     additional_forward_args=(layer_sizes, target_var, steps_or_samples), 
+                                                     target=target_class, n_samples=steps_or_samples)
+                        attributions = attributions.split(layer_sizes, dim = 3)
                     aggregated_attributions[target_class].append(attributions)
 
         # For each target class and for each data modality/layer, concatenate attributions accross batches 
